from a2ml.api.auger.impl.cloud.rest_api import RestApi
from a2ml.api.auger.impl.cloud.base import AugerBaseApi

ORGANIZATIONS = {
    'meta': {
        'status': 200,
        'pagination':
            {'limit': 100, 'total': 1, 'count': 1, 'offset': 0}
    },
    'data': [{'name': 'auger'}]
}

PROJECT_FILE = {
    'data': {
        'name': 'iris.csv',
        'id': 1256,
        'statistics': {
            'columns_count': 5, 'count': 150,
            'stat_data': [
            {
                'datatype': 'categorical',
                'column_name': 'species',
                'unique_values': 3
            },{
                'datatype': 'integer',
                'column_name': 'sepal_length'
            },{
                'datatype': 'integer',
                'column_name': 'sepal_width'
            },{
                'datatype': 'integer',
                'column_name': 'petal_length'
            },{
                'datatype': 'integer',
                'column_name': 'petal_width'
            }]
         },
    }
}

PROJECT_FILES = {
    'meta': {
        'pagination': {'offset': 0, 'count': 1, 'total': 1, 'limit': 100},
        'status': 200},
    'data': [PROJECT_FILE['data']]
}

EXPERIMENT = {
    'data': {
        'name': 'iris-1.csv-experiment',
        'project_file_id': 1256,
    }
}

EXPERIMENTS = {
    'meta': {
        'pagination': {'offset': 0, 'count': 1, 'total': 1, 'limit': 100},
        'status': 200},
    'data': [EXPERIMENT['data']]
}

PROJECTS = {
    'meta': {
        'status': 200,
        'pagination': {
            'count': 2, 'limit': 100, 'offset': 0, 'total': 2}},
    'data': [
        {"id": 2, "name": "project_1"},
        {"id": 1, "name": "cli-integration-test"}]
}

PAYLOAD_DEFAULT = {
    'get_organizations': ORGANIZATIONS,
    'get_projects': PROJECTS,
    'get_project_files': {
        'meta': {
            'status': 200,
            'pagination': {
                'count': 2, 'offset': 0, 'total': 2, 'limit': 100}
        },
        'data': [
        {'id': 1, 'name': 'test_dataset1'},
        {'id': 2, 'name': 'test_dataset2'}
        ],
    }
}

def interceptor(payload, monkeypatch):
    def payloader(x, method, *args, **kwargs):
        return payload[method]
    monkeypatch.setattr(
        RestApi, 'call_ex', payloader)


def object_status_chain(statuses, monkeypatch):
    current = statuses.pop(0)
    if len(statuses):
        monkeypatch.setattr(
            AugerBaseApi, 'wait_for_status', lambda x, *a, **kw: current)
    return current




import click
from a2ml.api.a2ml_project import A2MLProject
from a2ml.api.utils.context import pass_context

@click.group('project', short_help='Projects management')
@pass_context
def cmdl(ctx):
    """Project(s) management"""
    ctx.setup_logger(format='')

@click.command(short_help='List Projects')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def list_cmd(ctx, provider):
    """List Projects"""
    A2MLProject(ctx, provider).list()

@click.command(short_help='Create Project')
@click.argument('name', required=False, type=click.STRING)
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def create(ctx, provider, name):
    """Create Project"""
    A2MLProject(ctx, provider).create(name)

@click.command(short_help='Delete Project')
@click.argument('name', required=False, type=click.STRING)
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def delete(ctx, provider, name):
    """Delete Project"""
    A2MLProject(ctx, provider).delete(name)


@click.command(short_help='Select Project')
@click.argument('name', required=True, type=click.STRING)
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def select(ctx, provider, name):
    """Select Project.
       Name will be set in provider.yaml/project
    """
    A2MLProject(ctx, provider).select(name)


@pass_context
def add_commands(ctx):
    cmdl.add_command(list_cmd, name='list')
    cmdl.add_command(create)
    cmdl.add_command(delete)
    cmdl.add_command(select)


add_commands()



import click

from a2ml.api.utils.context import pass_context

@click.command('server', short_help='Run server')
@click.option('--host', '-h', type=click.STRING, required=False, default='0.0.0.0',
    show_default=True, help='host interface to bind server')
@click.option('--port', '-p', type=click.INT, required=False, default=8000,
    show_default=True, help='port to bind server')
@click.option('--reload', '-r', type=click.BOOL, required=False, default=False,
    show_default=True, help='enable autoreload')
@click.option('--workers', '-w', type=click.INT, required=False, default=1,
    show_default=True, help='workers count')
@pass_context
def cmdl(ctx, host, port, reload, workers):
    import uvicorn

    uvicorn.run(
        "a2ml.server.server:app",
        host=host, port=port, log_level="info", reload=reload, workers=workers
    )



import click

from a2ml.api.a2ml_credentials import A2MLCredentials
from a2ml.api.utils.context import pass_context

class AuthCmd(object):

    def __init__(self, ctx, provider='auger'):
        self.ctx = ctx
        self.provider = provider

    def login(self,username, password, organization, url):
        A2MLCredentials(self.ctx, self.provider).login(username, password, organization, url)

    def logout(self):
        A2MLCredentials(self.ctx, self.provider).logout()

    def whoami(self):
        A2MLCredentials(self.ctx, self.provider).whoami()

def prompt_login(ctx, param, provider):
    system = ctx.params.get('system','production')
    if provider == 'auger':
        username = ctx.params.get('username')
        if not username:
            username = click.prompt('username', default=None)

        organization = ctx.params.get('organization')
        if not organization:
            organization = click.prompt('organization', default=None)

        password = ctx.params.get('password')
        if not password:
            password = click.prompt('password', default=None, confirmation_prompt=False, hide_input=True)

        return {'username':username, 'organization':organization, 'password':password, 'system':system, 'name':provider}
    return {'username':None, 'organization':None, 'password':None, 'system':system, 'name':provider}

@click.group('auth', short_help='Authenticate with AutoML provider.')
@pass_context
def cmdl(ctx):
    """Authenticate with AutoML provider."""
    ctx.setup_logger(format='')

@click.command(short_help='Login to a provider.')
@click.option('--provider', '-p', default='auger', help='Provider name.', callback=prompt_login, show_default=True)
@click.option('--username', '-u', default=None, type=click.STRING, help='Provider username.', is_eager=True)
@click.option('--organization', '-o', default=None, type=click.STRING, help='Provider organization.', is_eager=True)
@click.option('--password', '-w', confirmation_prompt=False, default=None, help='Provider password.', is_eager=True, hide_input=True)
@click.option('--system', '-s', default='production',
    type=click.Choice(['production','staging']), help='Auger Provider endpoint.', show_default=True)
@pass_context
def login(ctx, username, password, organization, system, provider):
    """Login to a provider."""
    urls = {
        'production': 'https://app.auger.ai',
        'staging': 'https://app-staging.auger.ai'}
    AuthCmd(ctx, provider['name']).login(provider['username'], provider['password'], provider['organization'], urls[provider['system']])

@click.command(short_help='Logout from a provider.')
@click.option('--provider', '-p', default='auger', type=click.Choice(['auger','azure']), help='Provider name.', show_default=True)
@pass_context
def logout(ctx, provider):
    """Logout from a provider."""
    AuthCmd(ctx, provider).logout()

@click.command(short_help='Display the current logged in user for a provider.')
@click.option('--provider', '-p', default='auger', type=click.Choice(['auger','azure']), help='Provider name.', show_default=True)
@pass_context
def whoami(ctx, provider):
    """Display the current logged in user."""
    ctx.setup_logger(format='')
    AuthCmd(ctx, provider).whoami()


@pass_context
def add_commands(ctx):
    cmdl.add_command(login)
    cmdl.add_command(logout)
    cmdl.add_command(whoami)

add_commands()



import os
import time
import json
import requests
import shortuuid
import urllib.parse
import urllib.request
import xml.etree.ElementTree as ElementTree

from .cluster import AugerClusterApi
from .project_file import AugerProjectFileApi
from ..exceptions import AugerException
from .cluster_task import AugerClusterTaskApi

from  a2ml.api.utils import fsclient
from  a2ml.api.utils.file_uploader import FileUploader, NewlineProgressPercentage

SUPPORTED_FORMATS = ['.csv', '.arff', '.gz', '.bz2', '.zip', '.xz', '.json', '.xls', '.xlsx', '.feather', '.h5', '.hdf5', '.parquet']


class AugerDataSetApi(AugerProjectFileApi):
    """Auger DataSet API."""

    def __init__(self, ctx, project_api=None,
        data_set_name=None, data_set_id=None):
        super(AugerDataSetApi, self).__init__(
            ctx, project_api, data_set_name, data_set_id)

    def do_upload_file(self, data_source_file, data_set_name=None, local_data_source=True):
        # data_source_file, local_data_source = \
        #     AugerDataSetApi.verify(data_source_file, self.ctx.config.path)

        if local_data_source:
            file_url = self._upload_to_cloud(data_source_file)
            file_name = os.path.basename(data_source_file)
            if data_set_name:
                self.object_name = data_set_name
            else:
                self.object_name = self._get_data_set_name(file_name)
        else:
            file_url = data_source_file
            url_path = urllib.parse.urlparse(file_url).path
            file_name = os.path.basename(url_path)
            self.object_name = file_name

        return file_url, file_name

    def create(self, data_source_file, data_set_name=None, local_data_source=True):
        file_url, file_name = self.do_upload_file(data_source_file, data_set_name=data_set_name, local_data_source=local_data_source)

        try:
            return super().create(file_url, file_name)
        except Exception as exc:
            if 'en.errors.project_file.url_not_uniq' in str(exc):
                raise AugerException(
                    'DataSet already exists for %s' % file_url)
            raise

    def download(self, path_to_download):
        remote_file = self.properties().get('url')
        unused, ext = os.path.splitext(remote_file)
        filename, unused = os.path.splitext(self.name)
        local_file = os.path.abspath(
            os.path.join(path_to_download, filename+ext))

        remote_file = os.path.basename(remote_file)
        s3_signed_url = self.rest_api.call('get_project_file_url', {
            'project_id': self.parent_api.oid,
            'file_path': remote_file}).get('url')

        if not os.path.exists(path_to_download):
            os.makedirs(path_to_download)

        urllib.request.urlretrieve(s3_signed_url, local_file)

        return local_file

    def _get_readable_name(self):
        # patch readable name
        return 'DataSet'

    @staticmethod
    def verify(data_source_file, config_path=None):
        if urllib.parse.urlparse(data_source_file).scheme in ['http', 'https']:
            return data_source_file, False

        if not fsclient.is_s3_path(data_source_file):
            if config_path is None:
                config_path = os.getcwd()

            data_source_file = os.path.join(config_path, data_source_file)
            if not fsclient.is_s3_path(data_source_file):
                data_source_file = os.path.abspath(data_source_file)

        filename, file_extension = os.path.splitext(data_source_file)
        if not file_extension in SUPPORTED_FORMATS:
            raise AugerException(
                'Source file has to be one of the supported fomats: %s' %
                ', '.join(SUPPORTED_FORMATS))

        if not fsclient.is_file_exists(data_source_file):
            raise AugerException(
                'Can\'t find file to import: %s' % data_source_file)

        return data_source_file, True

    def _upload_to_cloud(self, file_to_upload):
        return self._upload_to_multi_tenant(file_to_upload)

    # def _upload_to_single_tenant(self, file_to_upload):
    #     # get file_uploader_service from the cluster
    #     # and upload data to that service
    #     project_properties = self.parent_api.properties()
    #     cluster_id = project_properties.get('cluster_id')
    #     cluster_api = AugerClusterApi(
    #         self.ctx, self.parent_api, cluster_id)
    #     cluster_properties = cluster_api.properties()

    #     file_uploader_service = cluster_properties.get('file_uploader_service')
    #     upload_token = file_uploader_service.get('params').get('auger_token')
    #     upload_url = '%s?auger_token=%s' % (
    #         file_uploader_service.get('url'), upload_token)

    #     file_url = self._upload_file(file_to_upload, upload_url)
    #     self.ctx.log(
    #         'Uploaded local file to Auger Cloud file: %s' % file_url)
    #     return file_url

    def _upload_file(self, file_name, url):
        with open(file_name, 'rb') as f:
            r = requests.post(url, data=f)

        if r.status_code == 200:
            rp = urllib.parse.parse_qs(r.text)
            return ('files/%s' % rp.get('path')[0].split('files/')[-1])
        else:
            raise AugerException(
                'HTTP error [%s] while uploading file to Auger Cloud...' % r.status_code)

    def _upload_to_multi_tenant(self, file_to_upload):
        file_path = 'workspace/projects/%s/files/%s-%s' % \
            (self.parent_api.object_name, shortuuid.uuid(),
             os.path.basename(file_to_upload))

        res = self.rest_api.call('create_project_file_url', {
            'project_id': self.parent_api.object_id,
            'file_path': file_path,
            'file_size': fsclient.get_file_size(file_to_upload),
            'async': True
        })
        cluster_task = AugerClusterTaskApi(self.ctx, cluster_task_id=res['id'])
        cluster_task.wait_for_status(['pending', 'received', 'started', 'retry'])
        res = cluster_task.properties().get('result')
        if not res:
            raise AugerException(
                'Error while uploading file to Auger Cloud...')

        if 'multipart' in res:
            upload_details = res['multipart']
            config = upload_details['config']

            uploader = FileUploader(
                upload_details['bucket'],
                config['endpoint'],
                config['access_key'],
                config['secret_key'],
                config['security_token']
            )

            with fsclient.open_file(file_to_upload, 'rb', encoding=None, auto_decompression=False) as f:
                return uploader.multipart_upload_obj(
                    f,
                    upload_details['key'],
                    callback=NewlineProgressPercentage(file_to_upload)
                )
        else:
            url = res['url']
            bucket = res['bucket']

            file_path = res['fields']['key']
            with fsclient.open_file(file_to_upload, 'rb', encoding=None, auto_decompression=False) as f:
                files = {'file': (file_path, f)}
                res = requests.post(url, data=res['fields'], files=files)

            if res.status_code == 201 or res.status_code == 200:
                return 's3://%s/%s' % (bucket, file_path)
            else:
                if res.status_code == 400 and b'EntityTooLarge' in res.content:
                    max_size = ElementTree.fromstring(res.content).find('MaxSizeAllowed').text
                    max_size_mb = int(max_size) / 1024 / 1024
                    raise AugerException('Data set size is limited to %.1f MB' % max_size_mb)
                else:
                    raise AugerException(
                        'HTTP error [%s] "%s" while uploading file'
                            ' to Auger Cloud...' % (res.status_code, res.content))

    def _get_data_set_name(self, filename):
        dot_index = filename.find('.')
        if dot_index>=0:
            fname = filename[:dot_index]
            fext = filename[dot_index:]
        else:
            fname = filename

        # fname, fext = os.path.splitext(file_name)
        return self._get_uniq_object_name(fname, "")



import click
from a2ml.api.a2ml import A2ML
from a2ml.api.utils.context import pass_context


@click.command('train', short_help='Train the model.')
@click.option('--provider', '-p', type=click.Choice(['auger','azure']), required=False,
    help='Cloud AutoML Provider.')
@pass_context
def cmdl(ctx, provider):
    """Train the model."""
    ctx.setup_logger(format='')
    A2ML(ctx, provider).train()



import time
from .base import AugerBaseApi
from .cluster import AugerClusterApi
from ..exceptions import AugerException


class AugerProjectApi(AugerBaseApi):
    """Auger Project API."""

    def __init__(self, ctx, org_api,
        project_name=None, project_id=None):
        super(AugerProjectApi, self).__init__(
            ctx, org_api, project_name, project_id)
        assert org_api is not None, 'Organization must be set for Project'
        self._set_api_request_path('AugerProjectApi')

    def is_running(self):
        return self.properties().get('status') == 'running'

    def create(self):
        return self._call_create({
            'name': self.object_name, 'organization_id': self.parent_api.oid})

    def start(self):
        project_properties = self.properties()
        if project_properties.get('status') != 'running':
            self.ctx.log('Starting Project...')
            self._do_start(project_properties)
            self.ctx.log('Started Project %s' % self.object_name)
        # else:
        #     self.ctx.log('Project is already running...')

        self._update_cluster_settings()
        
    def get_cluster_config(self, local_config = True):
        if local_config:
            return AugerClusterApi.get_cluster_settings(self.ctx)
        else:
            project_properties = self.properties()
            result = {
                'worker_type_id': project_properties.get('worker_type_id'),
                'workers_count': project_properties.get('workers_count')
            }
            if 'kubernetes_stack' in project_properties:
                result['kubernetes_stack'] = project_properties.get('kubernetes_stack')
                
            return result
                
    def update_cluster_config(self, params):
        remote_cluster = self.get_cluster_config(local_config=False)

        update_properties = {}
        props_to_update = ['worker_type_id', 'workers_count', 'kubernetes_stack']
        for prop in props_to_update:
            if remote_cluster.get(prop, params.get(prop)) != params.get(prop):
                update_properties[prop] = params.get(prop)

        if update_properties:
            self.ctx.log('Update project cluster: %s' % update_properties)            
            update_properties['id'] = self.object_id
            self._call_update(update_properties, progress=['undeployed', 'deployed', 'scaling', 'zero_scaled', 'deploying'])

        return True
            
    def _update_cluster_settings(self):
        local_cluster = self.get_cluster_config(local_config=True)
        self.update_cluster_config(local_cluster)
                
    def _do_start(self, project_properties):
        self._ensure_object_id()
        status = project_properties.get('status')

        project_status = ['deployed', 'deploying']
        if status in project_status:
            return self.wait_for_status(project_status)

        cluster_id = project_properties.get('cluster_id')
        cluster_api = AugerClusterApi(self.ctx, self, cluster_id)

        cluster_settings = cluster_api.get_cluster_settings(self.ctx)
        # self.rest_api.call('update_project', {
        #     'id': self.object_id,
        #     'cluster_autoterminate_minutes':
        #         cluster_settings.get('autoterminate_minutes')})
        try:
            self.rest_api.call('deploy_project', {
                'id': self.object_id,
                'worker_type_id': cluster_settings.get('worker_type_id'),
                'workers_count' : cluster_settings.get('workers_count'),
                'kubernetes_stack': cluster_settings.get('kubernetes_stack')})
        except:
            project_properties = self.properties()
            status = project_properties.get('status')
            project_status = ['deployed', 'deploying', 'running']
            if not status in project_status:
                raise                    

        result = self.wait_for_status(
            ['undeployed', 'deployed', 'scaling', 'zero_scaled', 'deploying'])
        time.sleep(20)
        return result

    def stop(self):
        if self.status() != 'undeployed':
            self.rest_api.call(
                'undeploy_project', {'id': self.object_id})
            return self.wait_for_status(['running', 'undeploying'])



import pytest
from a2ml.cmdl.cmdl import cmdl


class TestProjectCLI():
    def test_create_list_delete(self, runner, log, project):
        result = runner.invoke(cmdl, ['project', 'delete', 'cli-integration-test'])

        result = runner.invoke(cmdl, ['project', 'select', 'cli-integration-test'])
        assert log.messages[-1] == '[auger]  Selected Project cli-integration-test'

        result = runner.invoke(cmdl, ['project', 'create'])
        assert log.messages[-1] == '[auger]  Created Project cli-integration-test'

        result = runner.invoke(cmdl, ['project', 'list'])
        assert 'cli-integration-test' in str(log.messages)

        result = runner.invoke(cmdl, ['project', 'delete'])
        assert log.messages[-1] == '[auger]  Deleted Project cli-integration-test'
